/* memory.h
 *
 * Copyright (C) 2017 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */


#ifndef HAVE_MEMORY_H
#define HAVE_MEMORY_H

#include "xlib/xutil.h"


#define MEM_READ_WRITE       0
#define MEM_READ_ONLY        1

#define ADDRESS_SPACE        65536

#if 0
#define BYTE_ROTATE_LEFT(b)  (unsigned int)(((b) << 1) | ((b) >> 7))
#define BYTE_ROTATE_RIGHT(b) (unsigned int)(((b) >> 1) | ((b) << 7))
#endif

#define BYTE_HI(w)           (Byte)((w) >> 8)
#define BYTE_LO(w)           (Byte)(w)
#define WORD_BUILD(hi,lo)    (((hi) & 0xFF) << 8 | ((lo) & 0xFF))
#define ADDRESS(w)           ((w) & 0xFFFF)


typedef unsigned char Byte;
typedef unsigned short Word;


typedef enum {

	MEM_TYPE_RAM,
	MEM_TYPE_ROM,
	MEM_TYPE_VIDEO,
	MEM_TYPE_KEYBOARD,
	MEM_TYPE_IO,
	MEM_TYPE_ACIA

} MemoryType;


typedef struct _Memory Memory;

typedef void (*MemoryStoreByte_f) (Memory *m, int offset, Byte b);
typedef Byte (*MemoryReadByte_f)  (Memory *m, int offset);
typedef void (*MemoryFreeData_f)  (Memory *m);


struct _Memory {

	MemoryType type;

	int        base;
	int        size;
	xbool_t     read_only;

	MemoryFreeData_f  free_data;
	MemoryStoreByte_f store_byte;
	MemoryReadByte_f  read_byte;

	Byte       *data;

};


Memory* mem_new (int base, int size, MemoryType type);
void    mem_init (Memory* m, int base, int size, xbool_t randomize,
                  MemoryType type);
void    mem_free (Memory *m);

Byte    mem_read_byte (Memory *m, int offset);
void    mem_store_byte (Memory *m, int offset, Byte b);

void    mem_set_free_data_func (Memory *m, MemoryFreeData_f func);
void    mem_set_read_func (Memory *m, MemoryReadByte_f func);
void    mem_set_store_func (Memory *m, MemoryStoreByte_f func);

MemoryReadByte_f  mem_get_read_func (Memory *m);
MemoryStoreByte_f mem_get_store_func (Memory *m);

#ifdef DEBUG
void mem_set_quiet (xbool_t quiet);
#endif

#endif

