/* xtree.h
 *
 * Copyright (C) 2014, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */


#ifndef XTREE_H
#define XTREE_H


typedef xbool_t (*XTree_f)  (void *key, void *data, void *user);

typedef struct _xtree_t xtree_t;

#define xtree_contains(tree, key)  (xtree_lookup ((tree), (key)) != NULL)

xtree_t* xtree_new (XCmp_f cmp, XFree_f free_key, XFree_f free_data);
void     xtree_free (xtree_t *t);

void     xtree_insert (xtree_t *t, void *key, void *data);
void     xtree_replace (xtree_t *t, void *key, void *data);
void*    xtree_lookup (xtree_t *t, void *key);
void     xtree_remove (xtree_t *t, void *key);

int      xtree_n_items (xtree_t *t);
int      xtree_height (xtree_t *t);
int      xtree_shortest_path (xtree_t *t);

void     xtree_foreach (xtree_t *t, XTree_f func, void *user);


#endif /* XTREE_H */

